from Dictionary.Word cimport Word


cdef class TrieNode:

    cdef dict __children
    cdef set __words

    cpdef addWord(self, str word, Word root, int index=*)
    cpdef TrieNode getChild(self, char ch)
    cpdef set getWords(self)



cdef class TxtWord(Word):

    def __init__(self, name: str, flag=None):
        """
        Another constructor of TxtWord class which takes a String name and a flag as inputs and calls its super class
        Word with given name. Then, creates a new list as flags and calls addFlag method with given flag.

        PARAMETERS
        ----------
        name : str
            String input.
        flag : str
            String input.
        """
        super().__init__(name)
        self.__flags = []
        if flag is not None:
            self.addFlag(flag)

    cpdef addFlag(self, str flag):
        """
        The addFlag method takes a String flag as an input and adds given flag to the flags list.

        PARAMETERS
        ----------
        flag : str
            String input to add.
        """
        self.__flags.append(flag)

    cpdef removeFlag(self, str flag):
        """
        The removeFlag method takes a String flag as an input and removes given flag from the flags list.

        PARAMETERS
        ----------
        flag : str
            String input to remove.
        """
        self.__flags.remove(flag)

    cpdef str verbType(self):
        """
        The verbType method checks flags list and returns the corresponding cases.

        RETURNS
        -------
        str
            the corresponding cases.
        """
        if "F1P1-NO-REF" in self.__flags:
            """
            There is no example in the Turkish dictionary.
            """
            return "F1P1-NO-REF"
        if "F2P1-NO-REF" in self.__flags:
            """
            F2P1-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take PassiveHl suffix, can take CausativeT suffix. e.g. Doğ, göç, için
            """
            return "F2P1-NO-REF"
        if "F3P1-NO-REF" in self.__flags:
            """
            F3P1-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take Passive Hl suffix, can take CausativeT suffix. e.g. Ak
            """
            return "F3P1-NO-REF"
        if "F4P1-NO-REF" in self.__flags:
            """
            F4P1-NO-REF: The bare-form is a verb and depending on this attribute, the verb can't take PassiveHn 
            suffix, can take CausativeT suffix. e.g. Aksa
            """
            return "F4P1-NO-REF"
        if "F4PR-NO-REF" in self.__flags:
            """
            F4PR-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take PassiveHl suffix, can take CausativeT suffix. e.g. Çevir, göster
            """
            return "F4PR-NO-REF"
        if "F4PL-NO-REF" in self.__flags:
            """
            F4PL-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take CausativeT suffix. e.g. Azal, çoğal
            """
            return "F4PL-NO-REF"
        if "F4PW-NO-REF" in self.__flags:
            """
            F4PW-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take PassiveN suffix, can take CausativeT suffix. e.g. Birle, boya
            """
            return "F4PW-NO-REF"
        if "F5PL-NO-REF" in self.__flags:
            """
            F5PL-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take CausativeDHr suffix. e.g. Çal, kal
            """
            return "F5PL-NO-REF"
        if "F5PR-NO-REF" in self.__flags:
            """
            F5PR-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take PassiveHl suffix, can take CausativeDHr suffix. e.g. Birleş, çöz
            """
            return "F5PR-NO-REF"
        if "F5PW-NO-REF" in self.__flags:
            """
            F5PW-NO-REF: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn 
            suffix, can take PassiveHl suffix, can take CausativeDHr suffix. e.g. Ye
            """
            return "F5PW-NO-REF"
        if "F1P1" in self.__flags:
            """
            F1P1: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveHl suffix, can take CausativeAr suffix, can take ReciprocalHs suffix. e.g. Çık, kop
            """
            return "F1P1"
        if "F2P1" in self.__flags:
            """
            F2P1: The bare-form is a verb and depending on this attribute, the verb can can not PassiveHn suffix, can 
            take CausativeHr suffix, can take CausativeDHr suffix, can take ReciprocalHs suffix. e.g. Bit, doy, düş
            """
            return "F2P1"
        if "F2PL" in self.__flags:
            """
            F2PL: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take CausativeHr suffix, can take CausativeDHr suffix, can take ReciprocalHs suffix. e.g. Art, çök
            """
            return "F2PL"
        if "F3P1" in self.__flags:
            """
            F3P1: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveHl suffix, can take CausativeHl suffix, can take ReciprocalHs suffix. e.g. Kok, sark
            """
            return "F3P1"
        if "F4P1" in self.__flags:
            """
            F4P1: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix,
            can take CausativeT suffix, can take ReciprocalHs suffix. e.g. Anla
            """
            return "F4P1"
        if "F4PR" in self.__flags:
            """
            F4PR: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveHl suffix, can take CausativeT suffix, can take ReciprocalHs suffix. e.g. Bitir, çağır
            """
            return "F4PR"
        if "F4PL" in self.__flags:
            """
            F4PL: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveN suffix, can take CausativeT suffix, can take ReciprocalHs suffix. e.g. Bolal, çömel
            """
            return "F4PL"
        if "F4PW" in self.__flags:
            """
            F4PW: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveN suffix, can take CausativeT suffix, can take ReciprocalHs suffix. e.g. Boyla, çağla
            """
            return "F4PW"
        if "F5P1" in self.__flags:
            """
            F5P1: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveHl suffix, can take CausativeDHr suffix, can take ReciprocalHs suffix, can take ReflexiveHn 
            suffix. e.g. Giy
            """
            return "F5P1"
        if "F5PL" in self.__flags:
            """
            F5PL: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, can 
            take PassiveHl suffix, can take CausativeDHr suffix, can take ReciprocalHs suffix. e.g. Böl, dal
            """
            return "F5PL"
        if "F5PR" in self.__flags:
            """
            F5PR: The bare-form is a verb and depending on this attribute, the verb can take NominalVerb suffixes 
            "-sHm, -SHn, -yHz, SHnHz, -lAr", can take NominalVerb1 suffixes, "-yDH, -ysA", can take NominalVerb2 suffix,
            "-ymHs", can take AdjectiveRoot suffix, "-SH", can take Adjective suffix, "-ŞAr" e.g. Bilin, çalış
            """
            return "F5PR"
        if "F5PW" in self.__flags:
            """
            F5PW: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix,
            can take CausativeDHr suffix, can take ReciprocalHs suffix. e.g. Boşver, cezbet
            """
            return "F5PW"
        if "F6P1" in self.__flags:
            """
            F6P1: The bare-form is a verb and depending on this attribute, the verb can not take PassiveHn suffix, 
            can take PassiveN suffix, can take ReciprocalHs suffix, can take ReflexiveHn suffix. e.g. Gizle, hazırla, 
            kaşı
            """
            return "F6P1"
        return ""

    cpdef bint samePos(self, TxtWord word):
        """
        The samePos method takes TxtWord as input and returns true if;

        flags list contains CL_ISIM
        CL_ISIM: The bare-form of the word is a noun. e.g. Abla

        flags list contains CL_FIIL
        CL_FIIL: The bare-form of the word is a verb. e.g. Affet

        flags list contains IS_ADJ
        IS_ADJ: The bare-form of the word is a adjective. e.g. Acayip

        flags list contains IS_ZM
        IS_ZM: The bare-form of the word is a pronoun. e.g. Başkası

        flags list contains IS_ADVERB
        IS_ADVERB: The bare-form of the word is a adverb. e.g. Tekrar, açıktan, adeta

        PARAMETERS
        ----------
        word : TxtWord
            TxtWord type input.

        RETURNS
        -------
        bool
            true if given word is nominal, verb, adjective, pronoun or adverb, false otherwise.
        """
        if self.isNominal() and word.isNominal():
            return True
        if self.isVerb() and word.isVerb():
            return True
        if self.isAdjective() and word.isAdjective():
            return True
        if self.isPronoun() and word.isPronoun():
            return True
        if self.isAdverb() and word.isAdverb():
            return True
        return False

    cpdef bint isNominal(self):
        """
        The isNominal method returns true if flags list contains CL_ISIM.

        RETURNS
        -------
        bool
            true if flags list contains CL_ISIM.
        """
        return "CL_ISIM" in self.__flags

    cpdef bint isPassive(self):
        """
        The isPassive method returns true if flags list contains PASSIVE-HN.

        RETURNS
        -------
        bool
            true if flags list contains PASSIVE-HN.
        """
        return "PASSIVE-HN" in self.__flags

    cpdef bint isAbbreviation(self):
        """
        The isAbbreviation method returns true if flags list contains IS_KIS.

        RETURNS
        -------
        bool
            true if flags list contains IS_KIS.
        """
        return "IS_KIS" in self.__flags

    cpdef bint isInterjection(self):
        """
        The isInterjection method returns true if flags list contains IS_INTERJ.

        RETURNS
        -------
        bool
            true if flags list contains IS_INTERJ.
        """
        return "IS_INTERJ" in self.__flags

    cpdef bint isDuplicate(self):
        """
        The isDuplicate method returns true if flags list contains IS_DUP.

        RETURNS
        -------
        bool
            true if flags list contains IS_DUP.
        """
        return "IS_DUP" in self.__flags

    cpdef bint isHeader(self):
        """
        The isHeader method returns true if flags list contains IS_HEADER.

        RETURNS
        -------
        bool
            true if flags list contains IS_HEADER.
        """
        return "IS_HEADER" in self.__flags

    cpdef bint isAdjective(self):
        """
        The isAdjective method returns true if flags list contains IS_ADJ.

        RETURNS
        -------
        bool
            true if flags list contains IS_ADJ.
        """
        return "IS_ADJ" in self.__flags

    cpdef bint isPureAdjective(self):
        """
        The isPureAdjective method returns true if flags list contains IS_PUREADJ.

        RETURNS
        -------
        bool
            true if flags list contains IS_PUREADJ.
        """
        return "IS_PUREADJ" in self.__flags

    cpdef bint isPronoun(self):
        """
        The isPronoun method returns true if flags list contains IS_ZM.
        IS_ZM: The bare-form of the word is a pronoun. e.g. Hangi, hep, hiçbiri

        RETURNS
        -------
        bool
            true if flags list contains IS_ZM.
        """
        return "IS_ZM" in self.__flags

    cpdef bint isQuestion(self):
        """
        The isQuestion method returns true if flags list contains IS_QUES.
        IS_QUES: The bare-form of the word is a question. e.g. Mi, mu, mü

        RETURNS
        -------
        bool
            true if flags list contains IS_QUES.
        """
        return "IS_QUES" in self.__flags

    cpdef bint isVerb(self):
        """
        The isVerb method returns true if flags list contains CL_FIIL.

        RETURNS
        -------
        bool
            true if flags list contains CL_FIIL.
        """
        return "CL_FIIL" in self.__flags

    cpdef bint isPortmanteau(self):
        """
        The isPortmanteau method returns true if flags list contains IS_BILEŞ.
        IS_BILEŞ: The bare-form is a portmanteau word in affixed form. e.g. gelinçiçeği

        RETURNS
        -------
        bool
            true if flags list contains IS_BILEŞ.
        """
        return "IS_BILEŞ" in self.__flags

    cpdef bint isDeterminer(self):
        """
        The isDeterminer method returns true if flags list contains IS_DET.
        IS_DET: The bare-form of the word is a determiner. e.g. Bazı, bir

        RETURNS
        -------
        bool
            true if flags list contains IS_DET.
        """
        return "IS_DET" in self.__flags

    cpdef bint isConjunction(self):
        """
        The isConjunction method returns true if flags list contains IS_CONJ.
        IS_CONJ: The bare-form of the word is a conjunction. e.g. Gerek, halbuki

        RETURNS
        -------
        bool
            true if flags list contains IS_CONJ.
        """
        return "IS_CONJ" in self.__flags

    cpdef bint isAdverb(self):
        """
        The isAdverb method returns true if flags list contains IS_ADVERB.

        RETURNS
        -------
        bool
            true if flags list contains IS_ADVERB.
        """
        return "IS_ADVERB" in self.__flags

    cpdef bint isPostP(self):
        """
        The isPostP method returns true if flags list contains IS_POSTP.
        IS_POSTP: The bare-form of the word is a postposition. e.g. Önce, takdirde, üzere

        RETURNS
        -------
        bool
            true if flags list contains IS_POSTP.
        """
        return "IS_POSTP" in self.__flags

    cpdef bint isPortmanteauEndingWithSI(self):
        """
        The isPortmanteauEndingWithSI method returns true if flags list contains IS_B_SI.
        IS_B_SI: The bare-form is a portmanteau word ending with "sı". e.g. Giritlalesi

        RETURNS
        -------
        bool
            true if flags list contains IS_B_SI.
        """
        return "IS_B_SI" in self.__flags

    cpdef bint isPortmanteauFacedVowelEllipsis(self):
        """
        The isPortmanteauFacedVowelEllipsis method returns true if flags list contains IS_B_UD.
        IS_B_UD: The bare-form of the word includes vowel epenthesis,
        therefore the last inserted vowel drops during suffixation. e.g. İnsanoğlu

        RETURNS
        -------
        bool
            true if flags list contains IS_B_UD.
        """
        return "IS_B_UD" in self.__flags

    cpdef bint isPortmanteauFacedSoftening(self):
        """
        The isPortmanteauFacedSoftening method returns true if flags list contains IS_B_SD.
        IS_B_SD: The bare-form of the word includes softening. e.g. Çançiçeği

        RETURNS
        -------
        bool
            true if flags list contains IS_B_SD.
        """
        return "IS_B_SD" in self.__flags

    cpdef bint isSuffix(self):
        """
        The isSuffix method returns true if flags list contains EK.
        EK: This tag indicates complementary verbs. e.g. İdi, iken, imiş.

        RETURNS
        -------
        bool
            true if flags list contains EK.
        """
        return "EK" in self.__flags

    cpdef bint isProperNoun(self):
        """
        The isProperNoun method returns true if flags list contains IS_OA.
        IS_OA: The bare-form of the word is a proper noun. e.g. Abant, Beşiktaş

        RETURNS
        -------
        bool
            true if flags list contains IS_OA.
        """
        return "IS_OA" in self.__flags

    cpdef bint isPlural(self):
        """
        The isPlural method returns true if flags list contains IS_CA.
        IS_CA: The bare-form of the word is already in a plural form,
        therefore can not take plural suffixes such as "ler" or "lar". e.g. Buğdaygiller

        RETURNS
        -------
        bool
            true if flags list contains IS_CA.
        """
        return "IS_CA" in self.__flags

    cpdef bint isNumeral(self):
        """
        The isNumeral method returns true if flags list contains IS_SAYI.
        IS_SAYI: The word is a number. e.g. Dört

        RETURNS
        -------
        bool
            true if flags list contains IS_SAYI.
        """
        return "IS_SAYI" in self.__flags

    cpdef bint isReal(self):
        """
        The isReal method returns true if flags list contains IS_REELSAYI.

        RETURNS
        -------
        bool
            true if flags list contains IS_REELSAYI.
        """
        return "IS_REELSAYI" in self.__flags

    cpdef bint isFraction(self):
        """
        The isFraction method returns true if flags list contains IS_KESIR.

        RETURNS
        -------
        bool
            true if flags list contains IS_KESIR.
        """
        return "IS_KESIR" in self.__flags

    cpdef bint isTime(self):
        """
        The isTime method returns true if flags list contains IS_ZAMAN.

        RETURNS
        -------
        bool
            true if flags list contains IS_ZAMAN.
        """
        return "IS_ZAMAN" in self.__flags

    cpdef bint isDate(self):
        """
        The isDate method returns true if flags list contains IS_DATE.

        RETURNS
        -------
        bool
            true if flags list contains IS_DATE.
        """
        return "IS_DATE" in self.__flags

    cpdef bint isPercent(self):
        """
        The isPercent method returns true if flags list contains IS_PERCENT.

        RETURNS
        -------
        bool
            true if flags list contains IS_PERCENT.
        """
        return "IS_PERCENT" in self.__flags

    cpdef bint isRange(self):
        """
        The isRange method returns true if flags list contains IS_RANGE.

        RETURNS
        -------
        bool
            true if flags list contains IS_RANGE.
        """
        return "IS_RANGE" in self.__flags

    cpdef bint isOrdinal(self):
        """
        The isOrdinal method returns true if flags list contains IS_ORD.
        IS_ORD: The bare-form of the word can take suffixes and these suffixes define a ranking. e.g. Birinci

        RETURNS
        -------
        bool
            true if flags list contains IS_ORD.
        """
        return "IS_ORD" in self.__flags

    cpdef bint notObeysVowelHarmonyDuringAgglutination(self):
        """
        The notObeysVowelHarmonyDuringAgglutination method returns true if flags list contains IS_UU.
        IS_UU: The bare-form does not obey vowel harmony while taking suffixes. e.g. Dikkat

        RETURNS
        -------
        bool
            true if flags list contains IS_UU.
        """
        return "IS_UU" in self.__flags

    cpdef bint obeysAndNotObeysVowelHarmonyDuringAgglutination(self):
        """
        The obeysAndNotObeysVowelHarmonyDuringAgglutination method returns true if flags list contains IS_UUU.
        IS_UUU: The bare-form does not obey vowel harmony while taking suffixes. e.g. Bol, kalp

        RETURNS
        -------
        bool
            true if flags list contains IS_UUU.
        """
        return "IS_UUU" in self.__flags

    cpdef bint rootSoftenDuringSuffixation(self):
        """
        The rootSoftenDuringSuffixation method returns true if flags list contains IS_SD, F_SD.

        RETURNS
        -------
        bool
            true if flags list contains IS_SD, F_SD.
        """
        return "IS_SD" in self.__flags or "F_SD" in self.__flags

    cpdef bint rootSoftenAndNotSoftenDuringSuffixation(self):
        """
        The rootSoftenAndNotSoftenDuringSuffixation method returns true if flags list contains IS_SDD.
        IS_SDD: The bare-form final consonant can (or can not) get devoiced during vowel-initial suffixation. e.g. Kalp

        RETURNS
        -------
        bool
            true if flags list contains IS_SDD.
        """
        return "IS_SDD" in self.__flags

    cpdef bint verbSoftenDuringSuffixation(self):
        """
        The verbSoftenDuringSuffixation method returns true if flags list contains F_SD.
        F_SD: The bare-form final consonant gets devoiced during vowel-initial suffixation. e.g. Cezbet

        RETURNS
        -------
        bool
            true if flags list contains F_SD.
        """
        return "F_SD" in self.__flags

    cpdef bint nounSoftenDuringSuffixation(self):
        """
        The nounSoftenDuringSuffixation method returns true if flags list contains IS_SD.
        IS_SD: The bare-form final consonant already has an accusative suffix. e.g. Kabağı

        RETURNS
        -------
        bool
            true if flags list contains IS_SD.
        """
        return "IS_SD" in self.__flags

    cpdef bint endingKChangesIntoG(self):
        """
        The endingKChangesIntoG method returns true if flags list contains IS_KG.
        IS_KG: The bare-form includes vowel epenthesis, therefore the last inserted vowel drope
        during suffixation. e.g. Çelenk

        RETURNS
        -------
        bool
            true if flags list contains IS_KG.
        """
        return "IS_KG" in self.__flags

    cpdef bint isExceptional(self):
        """
        The isExceptional method returns true if flags list contains IS_EX.
        IS_EX: This tag defines exception words. e.g. Delikanlı

        RETURNS
        -------
        bool
            true if flags list contains IS_EX.
        """
        return "IS_EX" in self.__flags

    cpdef bint duplicatesDuringSuffixation(self):
        """
        The duplicatesDuringSuffixation method returns true if flags list contains IS_ST.
        IS_ST: The second consonant of the bare-form undergoes a resyllabification. e.g. His

        RETURNS
        -------
        bool
            true if flags list contains IS_ST.
        """
        return "IS_ST" in self.__flags

    cpdef bint duplicatesAndNotDuplicatesDuringSuffixation(self):
        """
        The duplicatesAndNotDuplicatesDuringSuffixation method returns true if flags list contains IS_STT.
        IS_STT: The second consonant of the bare-form undergoes a resyllabification. e.g. His

        RETURNS
        -------
        bool
            true if flags list contains IS_STT.
        """
        return "IS_STT" in self.__flags

    cpdef bint lastIdropsDuringSuffixation(self):
        """
        The lastIdropsDuringSuffixation method returns true if flags list contains IS_UD.
        IS_UD: The bare-form includes vowel epenthesis, therefore the last inserted vowel drops during suffixation.
        e.g. Boyun

        RETURNS
        -------
        bool
            true if flags list contains IS_UD.
        """
        return "IS_UD" in self.__flags

    cpdef bint lastIDropsAndNotDropDuringSuffixation(self):
        """
        The lastIDropsAndNotDropDuringSuffixation method returns true if flags list contains IS_UDD.
        IS_UDD: The bare-form includes vowel epenthesis, therefore the last inserted vowel can (or can not) drop during
        suffixation. e.g. Kadir

        RETURNS
        -------
        bool
            true if flags list contains IS_UDD.
        """
        return "IS_UDD" in self.__flags

    cpdef bint takesRelativeSuffixKi(self):
        """
        The takesRelativeSuffixKi method returns true if flags list contains IS_KI.
        IS_KI: The word can take a suffix such as "ki". e.g. Önce

        RETURNS
        -------
        bool
            true if flags list contains IS_KI.
        """
        return "IS_KI" in self.__flags

    cpdef bint takesRelativeSuffixKu(self):
        """
        The takesRelativeSuffixKu method returns true if flags list contains IS_KU.
        IS_KU: The word can take a suffix such as "kü". e.g. Bugün

        RETURNS
        -------
        bool
            true if flags list contains IS_KU.
        """
        return "IS_KU" in self.__flags

    cpdef bint consonantSMayInsertedDuringPossesiveSuffixation(self):
        """
        The consonantSMayInsertedDuringPossesiveSuffixation method returns true if flags list contains IS_SII.

        RETURNS
        -------
        bool
            true if flags list contains IS_SII.
        """
        return "IS_SII" in self.__flags

    cpdef bint lastIdropsDuringPassiveSuffixation(self):
        """
        The lastIdropsDuringPassiveSuffixation method returns true if flags list contains F_UD.
        F_UD: The bare-form includes vowel epenthesis, therefore the last "ı"
        drops during passive suffixation. e.g. Çağır

        RETURNS
        -------
        bool
            true if flags list contains F_UD.
        """
        return "F_UD" in self.__flags

    cpdef bint vowelAChangesToIDuringYSuffixation(self):
        """
        The vowelAChangesToIDuringYSuffixation method returns true if flags list contains F_GUD.
        F_GUD: The verb bare-form includes vowel reduction, the last vowel "a" of the bare-form is replaced with "ı"
        e.g. Buzağıla

        RETURNS
        -------
        bool
            true if flags list contains F_GUD.
        """
        return "F_GUD" in self.__flags

    cpdef bint vowelEChangesToIDuringYSuffixation(self):
        """
        The vowelEChangesToIDuringYSuffixation method returns true if flags list contains F_GUDO.
        F_GUDO: The verb bare-form includes viwel reduction, the last vowel "e" of the
        bare-form is replaced with "i". e.g. Ye

        RETURNS
        -------
        bool
            true if flags list contains F_GUDO.
        """
        return "F_GUDO" in self.__flags

    cpdef bint takesSuffixIRAsAorist(self):
        """
        The takesSuffixIRAsAorist method returns true if flags list contains F_GIR.
        F_GIR: The bare-form of the word takes "ir" suffix. e.g. Geç

        RETURNS
        -------
        bool
            true if flags list contains F_GIR.
        """
        return "F_GIR" not in self.__flags

    cpdef bint takesSuffixDIRAsFactitive(self):
        """
        The takesSuffixDIRAsFactitive method returns true if flags list contains F_DIR.
        F_DIR: The bare-form of the word takes "tır" suffix. e.g. Daral

        RETURNS
        -------
        bool
            true if flags list contains F_DIR.
        """
        return "F_DIR" not in self.__flags

    cpdef bint showsSuRegularities(self):
        """
        The showsSuRegularities method returns true if flags list contains IS_SU.

        RETURNS
        -------
        bool
            true if flags list contains IS_SU.
        """
        return "IS_SU" in self.__flags

    cpdef bint containsFlag(self, str flag):
        """
        The containsFlag method returns true if flags list contains flag.

        RETURNS
        -------
        bool
            true if flags list contains flag.
        """
        return flag in self.__flags

    def __str__(self) -> str:
        result = super().__str__()
        for flag in self.__flags:
            result = result + " " + flag
        return result



from Dictionary.Pos import Pos


cdef class ExceptionalWord(Word):

    def __init__(self, name: str, root: str, pos: Pos):
        """
        A constructor of ExceptionalWord class which takes a Pos as a  part of speech and two Strings; name
        and root as inputs. Then, calls its super class Word with given name and initialises root and pos variables
        with given inputs.

        PARAMETERS
        ----------
        name : str
            String input.
        root : str
            String input.
        pos : Pos
            Pos type input.
        """
        super().__init__(name)
        self.__root = root
        self.__pos = pos

    cpdef str getRoot(self):
        """
        Getter for the root variable.

        RETURNS
        -------
        str
            root variable.
        """
        return self.__root

    cpdef object getPos(self):
        """
        Getter for the pos variable.

        RETURNS
        -------
        Pos
            pos variable.
        """
        return self.__pos



from Dictionary.Word cimport Word


cdef class TxtWord(Word):

    cdef list __flags

    cpdef addFlag(self, str flag)
    cpdef removeFlag(self, str flag)
    cpdef str verbType(self)
    cpdef bint samePos(self, TxtWord word)
    cpdef bint isNominal(self)
    cpdef bint isPassive(self)
    cpdef bint isAbbreviation(self)
    cpdef bint isInterjection(self)
    cpdef bint isDuplicate(self)
    cpdef bint isHeader(self)
    cpdef bint isAdjective(self)
    cpdef bint isPureAdjective(self)
    cpdef bint isPronoun(self)
    cpdef bint isQuestion(self)
    cpdef bint isVerb(self)
    cpdef bint isPortmanteau(self)
    cpdef bint isDeterminer(self)
    cpdef bint isConjunction(self)
    cpdef bint isAdverb(self)
    cpdef bint isPostP(self)
    cpdef bint isPortmanteauEndingWithSI(self)
    cpdef bint isPortmanteauFacedVowelEllipsis(self)
    cpdef bint isPortmanteauFacedSoftening(self)
    cpdef bint isSuffix(self)
    cpdef bint isProperNoun(self)
    cpdef bint isPlural(self)
    cpdef bint isNumeral(self)
    cpdef bint isReal(self)
    cpdef bint isFraction(self)
    cpdef bint isTime(self)
    cpdef bint isDate(self)
    cpdef bint isPercent(self)
    cpdef bint isRange(self)
    cpdef bint isOrdinal(self)
    cpdef bint notObeysVowelHarmonyDuringAgglutination(self)
    cpdef bint obeysAndNotObeysVowelHarmonyDuringAgglutination(self)
    cpdef bint rootSoftenDuringSuffixation(self)
    cpdef bint rootSoftenAndNotSoftenDuringSuffixation(self)
    cpdef bint verbSoftenDuringSuffixation(self)
    cpdef bint nounSoftenDuringSuffixation(self)
    cpdef bint endingKChangesIntoG(self)
    cpdef bint isExceptional(self)
    cpdef bint duplicatesDuringSuffixation(self)
    cpdef bint duplicatesAndNotDuplicatesDuringSuffixation(self)
    cpdef bint lastIdropsDuringSuffixation(self)
    cpdef bint lastIDropsAndNotDropDuringSuffixation(self)
    cpdef bint takesRelativeSuffixKi(self)
    cpdef bint takesRelativeSuffixKu(self)
    cpdef bint consonantSMayInsertedDuringPossesiveSuffixation(self)
    cpdef bint lastIdropsDuringPassiveSuffixation(self)
    cpdef bint vowelAChangesToIDuringYSuffixation(self)
    cpdef bint vowelEChangesToIDuringYSuffixation(self)
    cpdef bint takesSuffixIRAsAorist(self)
    cpdef bint takesSuffixDIRAsFactitive(self)
    cpdef bint showsSuRegularities(self)
    cpdef bint containsFlag(self, str flag)



from Language.TurkishLanguage import TurkishLanguage
import re


cdef class Word:

    def __init__(self, name: str):
        """
        A constructor of Word class which gets a String name as an input and assigns to the name variable.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __str__(self) -> str:
        """
        The overridden __str__ method returns the name variable.

        RETURNS
        -------
        str
            the name variable.
        """
        return self.name

    def __lt__(self, other: Word):
        return self.name < other.name

    def __eq__(self, other: Word):
        return self.name == other.name

    cpdef int charCount(self):
        """
        The charCount method returns the length of name variable.

        RETURNS
        -------
        int
            the length of name variable.
        """
        return len(self.name)

    cpdef str getName(self):
        """
        Getter for the name variable.

        RETURNS
        -------
        str
            name variable.
        """
        return self.name

    cpdef setName(self, str name):
        """
        Setter for the name variable.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.name = name

    @staticmethod
    def isCapital(surfaceForm: str) -> bool:
        """
        The isCapital method takes a String surfaceForm as an input and returns true if the character at first index of
        surfaceForm is a capital letter, false otherwise.

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check the first character.

        RETURNS
        -------
        bool
            true if the character at first index of surfaceForm is a capital letter, false otherwise.
        """
        return surfaceForm[0] in TurkishLanguage.UPPERCASE_LETTERS

    @staticmethod
    def isPunctuationSymbol(surfaceForm: str) -> bool:
        """
        The isPunctuationSymbol method takes a String surfaceForm as an input and returns true if it is a punctuation,
        False otherwise.
        Grave accent : \u0060
        Left quotation mark : \u201C
        Right quotation mark : \u201D
        Left single quotation mark : \u2018
        Horizontal ellipsis : \u2026

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check.

        RETURNS
        -------
        bool
            true if it is a punctuation, false otherwise.
        """
        return surfaceForm == "." or surfaceForm == "..." or surfaceForm == "[" or surfaceForm == "]" or \
               surfaceForm == "\u2026" or surfaceForm == "%" or surfaceForm == "&" or surfaceForm == "=" or \
               surfaceForm == "\u0060\u0060" or surfaceForm == "\u0060" or surfaceForm == "''" or surfaceForm == "$" or \
               surfaceForm == "!" or surfaceForm == "?" or surfaceForm == "," or surfaceForm == ":" or \
               surfaceForm == "--" or surfaceForm == ";" or surfaceForm == "(" or surfaceForm == ")" or \
               surfaceForm == "'" or surfaceForm == "\"" or surfaceForm == "\u201C" or surfaceForm == "\u2018" or \
               surfaceForm == "\u201D" or surfaceForm == "…" or surfaceForm == "\u25CF" or surfaceForm == "/" or \
               surfaceForm == "-" or surfaceForm == "+" or surfaceForm == "-LRB-" or surfaceForm == "-RRB-" or \
               surfaceForm == "-LCB-" or surfaceForm == "-RCB-" or surfaceForm == "-LSB-" or surfaceForm == "-RSB-"

    @staticmethod
    def isHonorific(surfaceForm: str) -> bool:
        """
        The isHonorific method takes a String surfaceForm as an input and after converting it to lower case it returns
        true if it equals to "bay" or "bayan", false otherwise.

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check.

        RETURNS
        -------
        bool
            true if it equals to "bay" or "bayan", false otherwise.
        """
        cdef str lowerCase
        lowerCase = surfaceForm.lower()
        return lowerCase == "bay" or lowerCase == "bayan"

    @staticmethod
    def isOrganization(surfaceForm: str) -> bool:
        """
        The isOrganization method takes a String surfaceForm as an input and after converting it to lower case it
        returns true if it equals to "corp", "inc.", or "co.", and false otherwise.

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check.

        RETURNS
        -------
        bool
            true if it equals to "corp", "inc.", or "co.", and false otherwise.
        """
        cdef str lowerCase
        lowerCase = surfaceForm.lower()
        return lowerCase == "corp" or lowerCase == "inc." or lowerCase == "co."

    @staticmethod
    def isMoney(surfaceForm: str) -> bool:
        """
        The isMoney method takes a String surfaceForm as an input and after converting it to lower case it returns true
        if it equals to one of the dolar, sterlin, paunt, ons, ruble, mark, frank, yan, sent, yen' or $, and false
        otherwise.

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check.

        RETURNS
        -------
        bool
            true if it equals to one of the dolar, sterlin, paunt, ons, ruble, mark, frank, yan, sent, yen' or $, and
            false otherwise.
        """
        cdef str lowerCase
        lowerCase = surfaceForm.lower()
        return lowerCase.startswith("dolar") or lowerCase.startswith("sterlin") or lowerCase.startswith("paunt") or \
               lowerCase.startswith("ons") or lowerCase.startswith("ruble") or lowerCase.startswith("mark") or \
               lowerCase.startswith("frank") or lowerCase == "yen" or lowerCase.startswith("sent") or \
               lowerCase.startswith("cent") or lowerCase.startswith("yen'") or ("$" in lowerCase)

    cpdef bint isPunctuation(self):
        """
        The isPunctuation method without any argument, it checks name variable whether it is a punctuation or not and
        returns true if so.

        RETURNS
        -------
        bool
            true if name is a punctuation.
        """
        return self.name == "," or self.name == "." or self.name == "!" or self.name == "?" or self.name == ":" or \
               self.name == ";" or self.name == "\"" or self.name == "''" or self.name == "'" or self.name == "`" or \
               self.name == "``" or self.name == "..." or self.name == "-" or self.name == "--"

    @staticmethod
    def isTime(surfaceForm: str) -> bool:
        """
        The isTime method takes a String surfaceForm as an input and after converting it to lower case it checks some
        cases.
        If it is in the form of 12:23:34 or 12:23 it returns true.
        If it starts with name of months; ocak, şubat, mart, nisan, mayıs, haziran, temmuz, ağustos, eylül, ekim, kasım,
        aralık it returns true.
        If it equals to the name of days; pazar, pazartesi, salı, çarşamba, perşembe, cuma, cumartesi it returns true.

        PARAMETERS
        ----------
        surfaceForm : str
            String input to check.

        RETURNS
        -------
        bool
            true if it presents time, and false otherwise.
        """
        cdef str lowerCase
        cdef int time
        lowerCase = surfaceForm.lower()
        if re.fullmatch("(\\d\\d|\\d):(\\d\\d|\\d):(\\d\\d|\\d)", lowerCase) is not None or \
                re.fullmatch("(\\d\\d|\\d):(\\d\\d|\\d)", lowerCase) is not None:
            return True
        if lowerCase.startswith("ocak") or lowerCase.startswith("şubat") or lowerCase.startswith("mart") or \
                lowerCase.startswith("nisan") or lowerCase.startswith("mayıs") or lowerCase.startswith("haziran") \
                or lowerCase.startswith("temmuz") or lowerCase.startswith("ağustos") or lowerCase.startswith("eylül") \
                or lowerCase.startswith("ekim") or lowerCase.startswith("kasım") or lowerCase == "aralık":
            return True
        if lowerCase == "pazar" or lowerCase == "salı" or lowerCase.startswith("çarşamba") or \
                lowerCase.startswith("perşembe") or lowerCase == "cuma" or lowerCase.startswith("cumartesi") \
                or lowerCase.startswith("pazartesi"):
            return True
        if "'" in lowerCase:
            lowerCase = lowerCase[0:lowerCase.find("'")]
        try:
            time = int(lowerCase)
            if 1900 < time < 2200:
                return True
        except ValueError:
            return False
        return False

    @staticmethod
    def toWordArray(sourceArray: list)-> list:
        """
        The toWordArray method takes a String sourceArray as an input. First it creates a Word type result list and puts
        items of input sourceArray to this result list.

        PARAMETERS
        ----------
        sourceArray : list
            String list.

        RETURNS
        -------
        list
            Word type list.
        """
        cdef list result
        result = []
        for word in sourceArray:
            result.append(Word(word))
        return result

    cpdef list toCharacters(self):
        """
        The toCharacters method creates a Word type characters list and adds characters of name variable
        to newly created list.

        RETURNS
        -------
        list
            Word type list.
        """
        cdef list characters
        cdef int i
        characters = []
        for i in range(len(self.name)):
            characters.append(Word(self.name[i]))
        return characters



from Dictionary.Trie.Trie cimport Trie
from Dictionary.Dictionary cimport Dictionary
from Dictionary.TxtWord cimport TxtWord
from Dictionary.Word cimport Word


cdef class TxtDictionary(Dictionary):

    cdef dict __misspelledWords

    cpdef addNumber(self, str name)
    cpdef addRealNumber(self, str name)
    cpdef addFraction(self, str name)
    cpdef addTime(self, str name)
    cpdef bint addProperNoun(self, str name)
    cpdef bint addNoun(self, str name)
    cpdef bint addVerb(self, str name)
    cpdef bint addAdjective(self, str name)
    cpdef bint addAdverb(self, str name)
    cpdef bint addPronoun(self, str name)
    cpdef bint addWithFlag(self, str name, str flag)
    cpdef mergeDictionary(self, str secondFileName, str mergedFileName)
    cpdef __loadFromText(self, str fileName)
    cpdef __loadMisspelledWords(self, str fileName)
    cpdef str getCorrectForm(self, str misspelledWord)
    cpdef saveAsTxt(self, str fileName)
    cpdef __addWordWhenRootSoften(self, Trie trie, Py_UNICODE last, str root, TxtWord word)
    cpdef Trie prepareTrie(self)


import math
from Math.Vector cimport Vector


cdef class VectorizedDictionary(Dictionary):

    def __init__(self, comparator=None):
        """
        A constructor of VectorizedDictionary class which calls its super class Dictionary.
        """
        super().__init__(comparator)

    cpdef addWord(self, VectorizedWord word):
        """
        The addWord method takes a VectorizedWord as an input and adds it to the words list.

        PARAMETERS
        ----------
        word : VectorizedWord
            VectorizedWord input.
        """
        self.words.append(word)

    cpdef VectorizedWord mostSimilarWord(self, str name):
        """
        The mostSimilarWord method takes a String name as an input, declares a maxDistance as -1 and creates a
        VectorizedWord word by getting the given name from words list. Then, it loops through the words list and if the
        current word is not equal to given input it calculates the distance between current word and given word by using
        dot product and updates the maximum distance. It then returns the result VectorizedWord which holds the most
        similar word to the given word.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        VectorizedWord
            VectorizedWord type result which holds the most similar word to the given word.
        """
        cdef double distance, maxDistance
        cdef VectorizedWord result, word, currentWord
        maxDistance = -1
        result = None
        word = self.getWord(name)
        if word is None:
            return None
        for currentWord in self.words:
            if currentWord != word and isinstance(word, VectorizedWord):
                distance = word.getVector().dotProduct(currentWord.getVector())
                if distance > maxDistance:
                    maxDistance = distance
                    result = currentWord
        return result

    cdef makeComparator(self, VectorizedWord comparedWord):
        def compare(wordA, wordB):
            v = comparedWord.getVector()
            vA = wordA.getVector()
            vB = wordB.getVector()
            result1 = v.dotProduct(vB) / math.sqrt(v.dotProductWithSelf() * vB.dotProductWithSelf())
            result2 = v.dotProduct(vA) / math.sqrt(v.dotProductWithSelf() * vA.dotProductWithSelf())
            if result1 < result2:
                return -1
            elif result1 > result2:
                return 1
            else:
                return 0
        return compare

    cpdef list mostSimilarKWords(self, str name, int k):
        """
        The mostSimilarKWords method takes a String name and an integer k as inputs, and creates an list resultWords
        of type VectorizedWord and a VectorizedWord word by getting the given name from words list. Then, it loops
        through the words list and adds current word to the resultWords. It then sort resultWords list and if the size
        of the list is greater than given input k, it removes items from the ending. Then, it returns resultWords list.

        PARAMETERS
        ----------
        name : str
            String input.
        k : int
            Integer input.

        RETURNS
        -------
        list
            list result.
        """
        cdef list resultWords
        cdef VectorizedWord word, currentWord
        resultWords = []
        word = self.getWord(name)
        if word is None:
            return resultWords
        for currentWord in self.words:
            resultWords.append(currentWord)
        resultWords.sort(key=self.makeComparator(word))
        return resultWords[0: k]

    cpdef list kMeansClustering(self, int iteration, int k):
        """
        The kMeansClustering method takes an integer iteration and k as inputs. K-means clustering aims to partition n
        observations into k clusters in which each observation belongs to the cluster with the nearest mean.

        PARAMETERS
        ----------
        iteration : int
            Integer input.
        k : int
            Integer input.

        RETURNS
        -------
        list
            list result which holds the k-means clustered words.
        """
        cdef list result, means
        cdef Vector v
        cdef int vectorSize, i, j, maxClusterIndex
        cdef double maxClusterDistance
        cdef VectorizedWord vectorizedWord
        result = []
        means = []
        vectorSize = self.words[0].getVector().size()
        for i in range(k):
            result.append([])
            v = Vector()
            v.initAllSame(vectorSize, 0)
            means.append(v)
        for i in range(len(self.words)):
            result[i % k].append(self.words[i])
            means[i % k].add(self.words[i]).getVector()
        for i in range(k):
            means[i].divide(len(result[i]))
            means[i].divide(math.sqrt(means[i].dotProductWithSelf()))
        for i in range(iteration):
            for j in range(k):
                result[j].clear()
            for vectorizedWord in self.words:
                maxClusterDistance = means[0].dotProduct(vectorizedWord.getVector())
                maxClusterIndex = 0
                for j in range(1, k):
                    clusterDistance = means[j].dotProduct(vectorizedWord.getVector())
                    if clusterDistance > maxClusterDistance:
                        maxClusterDistance = clusterDistance
                        maxClusterIndex = j
                result[maxClusterIndex].append(vectorizedWord)
            for j in range(k):
                means[j].clear()
                for word in result[j]:
                    means[j].add(word.getVector())
                means[j].divide(len(result[j]))
                means[j].divide(math.sqrt(means[j].dotProductWithSelf()))
        return result



from Dictionary.Dictionary cimport Dictionary
from Dictionary.VectorizedWord cimport VectorizedWord


cdef class VectorizedDictionary(Dictionary):

    cpdef addWord(self, VectorizedWord word)
    cpdef VectorizedWord mostSimilarWord(self, str name)
    cpdef list mostSimilarKWords(self, str name, int k)
    cpdef list kMeansClustering(self, int iteration, int k)
    cdef makeComparator(self, VectorizedWord comparedWord)


from Dictionary.Trie.TrieNode cimport TrieNode
from Dictionary.Word cimport Word
from Dictionary.TxtWord cimport TxtWord


cdef class Trie:

    cdef TrieNode __rootNode

    cpdef addWord(self, str word, Word root)
    cpdef set getWordsWithPrefix(self, str surfaceForm)
    cpdef TxtWord getCompundWordStartingWith(self, str _hash)


from bisect import bisect_left
from functools import cmp_to_key

import pkg_resources

cdef class TxtDictionary(Dictionary):

    def __init__(self, fileName=None, misspelledFileName=None, comparator=None):
        """
        Constructor of TxtDictionary class which takes a String filename as input. And calls its super class
        Dictionary, assigns given filename input to the filename variable. Then, it calls loadFromText method with given
        filename.

        PARAMETERS
        ----------
        fileName : str
            String input.
        """
        super().__init__(comparator)
        self.__misspelledWords = {}
        if fileName is None:
            fileName = pkg_resources.resource_filename(__name__, 'data/turkish_dictionary.txt')
            self.__loadMisspelledWords(pkg_resources.resource_filename(__name__, 'data/turkish_misspellings.txt'))
        self.filename = fileName
        self.__loadFromText(self.filename)
        if misspelledFileName is not None:
            self.__loadMisspelledWords(misspelledFileName)

    cpdef addNumber(self, str name):
        """
        The addNumber method takes a String name and calls addWithFlag method with given name and IS_SAYI flag.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.addWithFlag(name, "IS_SAYI")

    cpdef addRealNumber(self, str name):
        """
        The addRealNumber method takes a String name and calls addWithFlag method with given name and IS_REELSAYI flag.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.addWithFlag(name, "IS_REELSAYI")

    cpdef addFraction(self, str name):
        """
        The addFraction method takes a String name and calls addWithFlag method with given name and IS_KESIR flag.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.addWithFlag(name, "IS_KESIR")

    cpdef addTime(self, str name):
        """
        The addTime method takes a String name and calls addWithFlag method with given name and IS_ZAMAN flag.

        PARAMETERS
        ----------
        name : str
            String input.
        """
        self.addWithFlag(name, "IS_ZAMAN")

    cpdef bint addProperNoun(self, str name):
        """
        The addProperNoun method takes a String name and calls addWithFlag method with given name and IS_OA flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "IS_OA")

    cpdef bint addNoun(self, str name):
        """
        The addNoun method takes a String name and calls addWithFlag method with given name and CL_ISIM flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "CL_ISIM")

    cpdef bint addVerb(self, str name):
        """
        The addVerb method takes a String name and calls addWithFlag method with given name and CL_FIIL flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "CL_FIIL")

    cpdef bint addAdjective(self, str name):
        """
        The addAdjective method takes a String name and calls addWithFlag method with given name and IS_ADJ flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "IS_ADJ")

    cpdef bint addAdverb(self, str name):
        """
        The addAdverb method takes a String name and calls addWithFlag method with given name and IS_ADVERB flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "IS_ADVERB")

    cpdef bint addPronoun(self, str name):
        """
        The addPronoun method takes a String name and calls addWithFlag method with given name and IS_ZM flag.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        return self.addWithFlag(name, "IS_ZM")

    cpdef bint addWithFlag(self, str name, str flag):
        """
        The addWithFlag method takes a String name and a flag as inputs. First it creates a TxtWord word, then if
        given name is not in words list it creates new TxtWord with given name and assigns it to
        the word and adds given flag to the word, it also add newly created word to the words list's index
        found by performing a binary search and return true at the end. If given name is in words list,
        it adds it the given flag to the word.

        PARAMETERS
        ----------
        name : str
            String input.
        flag : str
            String flag.

        RETURNS
        -------
        bool
            true if given name is in words list, false otherwise.
        """
        cdef TxtWord word
        cdef int middle
        if self.getWord(name.lower()) is None:
            word = TxtWord(name.lower())
            word.addFlag(flag)
            middle = bisect_left(self.words, word)
            self.words.insert(middle, word)
            return True
        else:
            word = self.getWord(name.lower())
            if isinstance(word, TxtWord) and not word.containsFlag(flag):
                word.addFlag(flag)
        return False

    cpdef mergeDictionary(self, str secondFileName, str mergedFileName):
        """
        The mergeDictionary method takes a String inputs; secondFileName and mergedFileName. It reads given files line
        by line and splits them according to spaces and write each word to file whichever comes first lexicographically
        and continue to read files till the end.

        PARAMETERS
        ----------
        secondFileName : str
            String input.
        mergedFileName : str
            String input.
        """
        cdef int i, j
        cdef str st1, st2, word1, word2, flag
        cdef list firstLines, secondLines
        firstFile = open(self.filename, "r", encoding="utf8")
        secondFile = open(secondFileName, "r", encoding="utf8")
        outFile = open(mergedFileName, "w", encoding="utf8")
        firstLines = firstFile.readlines()
        firstFile.close()
        secondLines = secondFile.readlines()
        secondFile.close()
        i = 0
        j = 0
        while i < len(firstLines) and j < len(secondLines):
            st1 = firstLines[i]
            st2 = secondLines[j]
            word1 = st1.split()[0]
            word2 = st2.split()[0]
            if word1 < word2:
                print(st1, file=outFile)
                i = i + 1
            else:
                if word1 > word2:
                    print(st2, file=outFile)
                    j = j + 1
                else:
                    flag = st2.split()[1]
                    if flag in st1:
                        print(st1, file=outFile)
                    else:
                        print(st1 + " " + flag, file=outFile)
                    i = i + 1
                    j = j + 1
        while i < len(firstLines):
            st1 = firstLines[j]
            print(st1, file=outFile)
            i = i + 1
        while j < len(secondLines):
            st2 = secondLines[j]
            print(st2, file=outFile)
            j = j + 1
        outFile.close()

    cpdef __loadFromText(self, str fileName):
        """
        The loadFromText method takes a String filename as an input. It reads given file line by line and splits
        according to space and assigns each word to the String array. Then, adds these word with their flags to the
        words list. At the end it sorts the words list.

        PARAMETERS
        ----------
        fileName : str
            File name input.
        """
        cdef list lines, wordList
        cdef TxtWord currentWord
        cdef int i
        cdef str line
        inputFile = open(fileName, "r", encoding="utf8")
        lines = inputFile.readlines()
        for line in lines:
            wordList = line.split()
            if len(wordList) > 0:
                currentWord = TxtWord(wordList[0])
                for i in range(1, len(wordList)):
                    currentWord.addFlag(wordList[i])
                self.words.append(currentWord)
        inputFile.close()
        self.words.sort(key=cmp_to_key(self.comparator))

    cpdef __loadMisspelledWords(self, str fileName):
        """
        The loadMisspellWords method takes a String filename as an input. It reads given file line by line and splits
        according to space and assigns each word with its misspelled form to the the misspelledWords hashMap.

        PARAMETERS
        ----------
        fileName : str
            File name input.
        """
        cdef str line
        cdef list lines, wordList
        inputFile = open(fileName, "r", encoding="utf8")
        lines = inputFile.readlines()
        for line in lines:
            wordList = line.split()
            if len(wordList) == 2:
                self.__misspelledWords[wordList[0]] = wordList[1]
        inputFile.close()

    cpdef str getCorrectForm(self, str misspelledWord):
        """
        The getCorrectForm returns the correct form of a misspelled word.

        PARAMETERS
        ----------
        misspelledWord : str
            Misspelled form.

        RETURNS
        -------
        str
            Correct form.
        """
        if misspelledWord in self.__misspelledWords:
            return self.__misspelledWords[misspelledWord]
        return ""

    cpdef saveAsTxt(self, str fileName):
        """
        The saveAsTxt method takes a filename as an input and prints out the items of words list.

        PARAMETERS
        ----------
        fileName : str
            String input.
        """
        output = open(fileName, "w", encoding="utf8")
        for word in self.words:
            print(word, file=output)
        output.close()

    cpdef __addWordWhenRootSoften(self, Trie trie, Py_UNICODE last, str root, TxtWord word):
        """
        The addWordWhenRootSoften is used to add word to Trie whose last consonant will be soften.
        For instance, in the case of Dative Case Suffix, the word is 'müzik' when '-e' is added to the word, the last
        char is drooped and root became 'müzi' and by changing 'k' into 'ğ' the word transformed into 'müziğe' as in the
        example of 'Herkes müziğe doğru geldi'.

        In the case of accusative, possessive of third person and a derivative suffix, the word is 'kanat' when '-i' is
        added to word, last char is dropped, root became 'kana' then 't' transformed into 'd' and added to Trie. The
        word is changed into 'kanadı' as in the case of 'Kuşun kırık kanadı'.

        PARAMETERS
        ----------
        trie : Trie
            the name of the Trie to add the word.
        last : chr
            the last char of the word to be soften.
        root : str
            the substring of the word whose last one or two chars are omitted from the word to bo softed.
        word : TxtWord
            the original word.
        """
        if last == 'p':
            trie.addWord(root + 'b', word)
        elif last == 'ç':
            trie.addWord(root + 'c', word)
        elif last == 't':
            trie.addWord(root + 'd', word)
        elif last == 'k' or last == 'g':
            trie.addWord(root + 'ğ', word)
        else:
            pass

    cpdef Trie prepareTrie(self):
        """
        The prepareTrie method is used to create a Trie with the given dictionary. First, it gets the word from
        dictionary, then checks some exceptions like 'ben' which does not fit in the consonant softening rule and
        transforms into 'bana', and later on it generates a root by removing the last char from the word however if the
        length of the word is greater than 1, it also generates the root by removing the last two chars from the word.

        Then, it gets the last char of the root and adds root and word to the result Trie. There are also special cases
        such as;
        lastIdropsDuringSuffixation condition, if it is true then addWordWhenRootSoften method will be used rather than
        addWord.
        Ex : metin + i = metni
        isPortmanteauEndingWithSI condition, if it is true then addWord method with rootWithoutLastTwo will be used.
        Ex : ademelması + lar = ademelmaları
        isPortmanteau condition, if it is true then addWord method with rootWithoutLast will be used.
        Ex : mısıryağı + lar = mısıryağları
        vowelEChangesToIDuringYSuffixation condition, if it is then addWord method with rootWithoutLast will be used
        depending on the last char whether it is 'e' or 'a'.
        Ex : ye + iniz - yiyiniz
        endingKChangesIntoG condition, if it is true then addWord method with rootWithoutLast will be used with added
        'g'.
        Ex : ahenk + i = ahengi

        RETURNS
        -------
        Trie
            the resulting Trie.
        """
        cdef Trie result
        cdef int i, length
        cdef Word word
        cdef str root, rootWithoutLast, rootWithoutLastTwo
        cdef Py_UNICODE lastBefore
        result = Trie()
        lastBefore = ' '
        for i in range(self.size()):
            word = self.getWordWithIndex(i)
            if isinstance(word, TxtWord):
                root = word.getName()
                length = len(root)
                if root == "ben":
                    result.addWord("bana", word)
                if root == "sen":
                    result.addWord("sana", word)
                rootWithoutLast = root[0:length - 1]
                if length > 1:
                    rootWithoutLastTwo = root[0:length - 2]
                else:
                    rootWithoutLastTwo = ""
                if length > 1:
                    lastBefore = root[length - 2]
                last = root[length - 1]
                result.addWord(root, word)
                if word.lastIdropsDuringSuffixation() or word.lastIdropsDuringPassiveSuffixation():
                    if word.rootSoftenDuringSuffixation():
                        self.__addWordWhenRootSoften(result, last, rootWithoutLastTwo, word)
                    else:
                        result.addWord(rootWithoutLastTwo + last, word)
                if word.isPortmanteauEndingWithSI():
                    result.addWord(rootWithoutLastTwo, word)
                if word.rootSoftenDuringSuffixation():
                    self.__addWordWhenRootSoften(result, last, rootWithoutLast, word)
                if word.isPortmanteau():
                    if word.isPortmanteauFacedVowelEllipsis():
                        result.addWord(rootWithoutLastTwo + last + lastBefore, word)
                    else:
                        if word.isPortmanteauFacedSoftening():
                            if lastBefore == 'b':
                                result.addWord(rootWithoutLastTwo + 'p', word)
                            elif lastBefore == 'c':
                                result.addWord(rootWithoutLastTwo + 'ç', word)
                            elif lastBefore == 'd':
                                result.addWord(rootWithoutLastTwo + 't', word)
                            elif lastBefore == 'ğ':
                                result.addWord(rootWithoutLastTwo + 'k', word)
                            else:
                                pass
                        else:
                            result.addWord(rootWithoutLast, word)
                if word.vowelEChangesToIDuringYSuffixation() or word.vowelAChangesToIDuringYSuffixation():
                    if last == 'e':
                        result.addWord(rootWithoutLast, word)
                    elif last == 'a':
                        result.addWord(rootWithoutLast, word)
                    else:
                        pass
                if word.endingKChangesIntoG():
                    result.addWord(rootWithoutLast + 'g', word)
        return result



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
from Cython.Build import cythonize

setup(
    ext_modules=cythonize(["Dictionary/*.pyx", "Dictionary/Trie/*.pyx", "Language/*.pyx", "Syllibification/*.pyx"],
                          compiler_directives={'language_level': "3"}),
    name='NlpToolkit-Dictionary-Cy',
    version='1.0.13',
    packages=['Language', 'Dictionary', 'Dictionary.data', 'Dictionary.Trie', 'Syllibification'],
    package_data={'Language': ['*.pxd', '*.pyx', '*.c'],
                  'Dictionary': ['*.pxd', '*.pyx', '*.c', '*.py'],
                  'Dictionary.Trie': ['*.pxd', '*.pyx', '*.c'],
                  'Syllibification': ['*.pxd', '*.pyx', '*.c'],
                  'Dictionary.data': ['*.txt']},
    url='https://github.com/StarlangSoftware/Dictionary-Cy',
    license='',
    author='olcaytaner',
    author_email='olcay.yildiz@ozyegin.edu.tr',
    description='Simple Dictionary Processing',
    install_requires=['NlpToolkit-Math-Cy'],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



from Dictionary.Word cimport Word


cdef class Dictionary:

    def __init__(self, comparator=None):
        """
        An empty constructor of Dictionary class.
        """
        self.words = []
        self.filename = ""
        if comparator is None:
            self.comparator = Dictionary.turkishLowerCaseComparator
        else:
            self.comparator = comparator

    @staticmethod
    def turkishLowerCaseComparator(wordA: Word, wordB: Word):
        cdef int i, first, second
        cdef str firstChar, secondChar
        LOWERCASE_LETTERS = "abcçdefgğhıijklmnoöprsştuüvyz"
        for i in range(min(len(wordA.getName()), len(wordB.getName()))):
            firstChar = wordA.getName()[i:i + 1]
            secondChar = wordB.getName()[i:i + 1]
            if firstChar != secondChar:
                if firstChar in LOWERCASE_LETTERS and secondChar not in LOWERCASE_LETTERS:
                    return -1
                elif firstChar not in LOWERCASE_LETTERS and secondChar in LOWERCASE_LETTERS:
                    return 1
                elif firstChar in LOWERCASE_LETTERS and secondChar in LOWERCASE_LETTERS:
                    first = LOWERCASE_LETTERS.index(firstChar)
                    second = LOWERCASE_LETTERS.index(secondChar)
                    if first < second:
                        return -1
                    elif first > second:
                        return 1
        if len(wordA.getName()) < len(wordB.getName()):
            return -1
        elif len(wordA.getName()) > len(wordB.getName()):
            return 1
        else:
            return 0

    @staticmethod
    def turkishIgnoreCaseComparator(wordA: Word, wordB: Word):
        cdef int i, first, second
        cdef str firstChar, secondChar
        IGNORE_CASE_LETTERS = "aAbBcCçÇdDeEfFgGğĞhHıIiİjJkKlLmMnNoOöÖpPrRsSşŞtTuUüÜvVyYzZ"
        for i in range(min(len(wordA.getName()), len(wordB.getName()))):
            firstChar = wordA.getName()[i:i + 1]
            secondChar = wordB.getName()[i:i + 1]
            if firstChar != secondChar:
                if firstChar in IGNORE_CASE_LETTERS and secondChar not in IGNORE_CASE_LETTERS:
                    return -1
                elif firstChar not in IGNORE_CASE_LETTERS and secondChar in IGNORE_CASE_LETTERS:
                    return 1
                elif firstChar in IGNORE_CASE_LETTERS and secondChar in IGNORE_CASE_LETTERS:
                    first = IGNORE_CASE_LETTERS.index(firstChar)
                    second = IGNORE_CASE_LETTERS.index(secondChar)
                    if first < second:
                        return -1
                    elif first > second:
                        return 1
        if len(wordA.getName()) < len(wordB.getName()):
            return -1
        elif len(wordA.getName()) > len(wordB.getName()):
            return 1
        else:
            return 0

    cpdef Word getWord(self, str name):
        """
        The getWord method takes a String name as an input and performs binary search within words list and assigns the
        result to integer variable middle. If the middle is greater than 0, it returns the item at index middle of words
        list, None otherwise.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        Word
            the item at found index of words {@link ArrayList}, null if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(name)
        middle = self.__getPosition(word)
        if middle >= 0:
            return self.words[middle]
        return None

    cpdef int getWordIndex(self, str name):
        """
        The getWordIndex method takes a String name as an input and performs binary search within words list and assigns
        the result to integer variable middle. If the middle is greater than 0, it returns the index middle, -1
        otherwise.

        PARAMETERS
        ----------
        name : str
            String input.

        RETURNS
        -------
        int
            found index of words list, -1 if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(name)
        middle = self.__getPosition(word)
        if middle >= 0:
            return middle
        return -1

    cpdef removeWord(self, str name):
        """
        RemoveWord removes a word with the given name

        PARAMETERS
        ----------
        name : str
            Name of the word to be removed.
        """
        cdef int index
        index = self.getWordIndex(name)
        if index != -1:
            self.words.pop(index)

    cpdef int size(self):
        """
        The size method returns the size of the words list.

        RETURNS
        -------
        int
            The size of the words list.
        """
        return len(self.words)

    cpdef Word getWordWithIndex(self, int index):
        """
        The getWordWithIndex method which takes an index as an input and returns the value at given index of words list.

        PARAMETERS
        ----------
        index : int
            index to get the value.

        RETURNS
        -------
        Word
            The value at given index of words list.
        """
        return self.words[index]

    cpdef int longestWordSize(self):
        """
        The longestWordSize method loops through the words list and returns the item with the maximum word length.

        RETURNS
        -------
        int
            The item with the maximum word length.
        """
        cdef int maxLength
        maxLength = 0
        for word in self.words:
            if len(word.getName()) > maxLength:
                maxLength = len(word.getName())
        return maxLength

    cpdef int __getPosition(self, Word word):
        cdef int lo, hi, mid
        lo = 0
        hi = len(self.words) - 1
        while lo <= hi:
            mid = (lo + hi) // 2
            if self.comparator(self.words[mid], word) < 0:
                lo = mid + 1
            elif self.comparator(self.words[mid], word) > 0:
                hi = mid - 1
            else:
                return mid
        return -lo

    cpdef int getWordStartingWith(self, str _hash):
        """
        The getWordStartingWith method takes a String hash as an input and performs binary search within words list and
        assigns the result to integer variable middle. If the middle is greater than 0, it returns the index middle,
        -middle-1 otherwise.

        PARAMETERS
        ----------
        _hash : str
            String input.

        RETURNS
        -------
        int
            Found index of words list, -middle-1 if cannot be found.
        """
        cdef Word word
        cdef int middle
        word = Word(_hash)
        middle = self.__getPosition(word)
        if middle < 0:
            return -middle
        else:
            return middle
